@@ -11,8 +11,13 @@
 # FOR A PARTICULAR PURPOSE
 #
 ##############################################################################
+import mock
 import unittest
 
+import zope.interface.verify
+
+from .. import interfaces
+
 
 class TransactionManagerTests(unittest.TestCase):
 
@@ -31,6 +36,10 @@ class TransactionManagerTests(unittest.TestCase):
         nosub1 = DataObject(mgr, nost=1)
         return mgr, sub1, sub2, sub3, nosub1
 
+    def test_interface(self):
+        zope.interface.verify.verifyObject(interfaces.ITransactionManager,
+                                           self._makeOne())
+
     def test_ctor(self):
         tm = self._makeOne()
         self.assertTrue(tm._txn is None)
@@ -72,10 +81,10 @@ class TransactionManagerTests(unittest.TestCase):
         class Existing(object):
             _aborted = False
             def abort(self):
-                self._aborted = True
+                raise AssertionError("This is not actually called")
         tm = self._makeOne()
         tm._txn = txn = Existing()
-        self.assertTrue(tm.get() is txn)
+        self.assertIs(tm.get(), txn)
 
     def test_free_w_other_txn(self):
         from transaction._transaction import Transaction
@@ -88,11 +97,11 @@ class TransactionManagerTests(unittest.TestCase):
         class Existing(object):
             _aborted = False
             def abort(self):
-                self._aborted = True
+                raise AssertionError("This is not actually called")
         tm = self._makeOne()
         tm._txn = txn = Existing()
         tm.free(txn)
-        self.assertTrue(tm._txn is None)
+        self.assertIsNone(tm._txn)
 
     def test_registerSynch(self):
         tm = self._makeOne()
@@ -105,12 +114,27 @@ class TransactionManagerTests(unittest.TestCase):
         tm = self._makeOne()
         synch1 = DummySynch()
         synch2 = DummySynch()
+        self.assertFalse(tm.registeredSynchs())
         tm.registerSynch(synch1)
+        self.assertTrue(tm.registeredSynchs())
         tm.registerSynch(synch2)
+        self.assertTrue(tm.registeredSynchs())
         tm.unregisterSynch(synch1)
+        self.assertTrue(tm.registeredSynchs())
         self.assertEqual(len(tm._synchs), 1)
         self.assertFalse(synch1 in tm._synchs)
         self.assertTrue(synch2 in tm._synchs)
+        tm.unregisterSynch(synch2)
+        self.assertFalse(tm.registeredSynchs())
+
+    def test_clearSynchs(self):
+        tm = self._makeOne()
+        synch1 = DummySynch()
+        synch2 = DummySynch()
+        tm.registerSynch(synch1)
+        tm.registerSynch(synch2)
+        tm.clearSynchs()
+        self.assertEqual(len(tm._synchs), 0)
 
     def test_isDoomed_wo_existing_txn(self):
         tm = self._makeOne()
@@ -164,7 +188,7 @@ class TransactionManagerTests(unittest.TestCase):
             def commit(self):
                 self._committed = True
             def abort(self):
-                self._aborted = True
+                raise AssertionError("This should not be called")
         tm = self._makeOne()
         with tm:
             tm._txn = txn = _Test()
@@ -176,16 +200,16 @@ class TransactionManagerTests(unittest.TestCase):
             _committed = False
             _aborted = False
             def commit(self):
-                self._committed = True
+                raise AssertionError("This should not be called")
             def abort(self):
                 self._aborted = True
         tm = self._makeOne()
-        try:
+
+        with self.assertRaises(ZeroDivisionError):
             with tm:
                 tm._txn = txn = _Test()
-                1/0
-        except ZeroDivisionError: 
-            pass
+                raise ZeroDivisionError()
+
         self.assertFalse(txn._committed)
         self.assertTrue(txn._aborted)
 
@@ -221,6 +245,79 @@ class TransactionManagerTests(unittest.TestCase):
         self.assertEqual(len(found), 1)
         self.assertTrue(found[0] is tm)
 
+    def test_attempts_stop_on_success(self):
+        tm = self._makeOne()
+
+        i = 0
+        for attempt in tm.attempts():
+            with attempt:
+                i += 1
+
+        self.assertEqual(i, 1)
+
+    def test_attempts_retries(self):
+        import transaction.interfaces
+        class Retry(transaction.interfaces.TransientError):
+            pass
+
+        tm = self._makeOne()
+        i = 0
+        for attempt in tm.attempts(4):
+            with attempt:
+                i += 1
+                if i < 4:
+                    raise Retry
+
+        self.assertEqual(i, 4)
+
+    def test_attempts_retries_but_gives_up(self):
+        import transaction.interfaces
+        class Retry(transaction.interfaces.TransientError):
+            pass
+
+        tm = self._makeOne()
+        i = 0
+
+        with self.assertRaises(Retry):
+            for attempt in tm.attempts(4):
+                with attempt:
+                    i += 1
+                    raise Retry
+
+        self.assertEqual(i, 4)
+
+    def test_attempts_propigates_errors(self):
+        tm = self._makeOne()
+        with self.assertRaises(ValueError):
+            for attempt in tm.attempts(4):
+                with attempt:
+                    raise ValueError
+
+    def test_attempts_defer_to_dm(self):
+        import transaction.tests.savepointsample
+
+        class DM(transaction.tests.savepointsample.SampleSavepointDataManager):
+            def should_retry(self, e):
+                if 'should retry' in str(e):
+                    return True
+
+        ntry = 0
+        dm = transaction.tests.savepointsample.SampleSavepointDataManager()
+        dm2 = DM()
+        with transaction.manager:
+            dm2['ntry'] = 0
+
+        for attempt in transaction.manager.attempts():
+            with attempt:
+                ntry += 1
+                dm['ntry'] = ntry
+                dm2['ntry'] = ntry
+                if ntry % 3:
+                    raise ValueError('we really should retry this')
+
+        self.assertEqual(ntry, 3)
+
+
     def test_attempts_w_default_count(self):
         from transaction._manager import Attempt
         tm = self._makeOne()
@@ -231,6 +328,133 @@ class TransactionManagerTests(unittest.TestCase):
             self.assertTrue(attempt.manager is tm)
         self.assertTrue(found[-1] is tm)
 
+    def test_run(self):
+        import transaction.interfaces
+        class Retry(transaction.interfaces.TransientError):
+            pass
+
+        tm = self._makeOne()
+        i = [0, None]
+
+        @tm.run()
+        def meaning():
+            "Nice doc"
+            i[0] += 1
+            i[1] = tm.get()
+            if i[0] < 3:
+                raise Retry
+            return 42
+
+        self.assertEqual(i[0], 3)
+        self.assertEqual(meaning, 42)
+        self.assertEqual(i[1].description, "meaning\n\nNice doc")
+
+    def test_run_no_name_explicit_tries(self):
+        import transaction.interfaces
+        class Retry(transaction.interfaces.TransientError):
+            pass
+
+        tm = self._makeOne()
+        i = [0, None]
+
+        @tm.run(4)
+        def _():
+            "Nice doc"
+            i[0] += 1
+            i[1] = tm.get()
+            if i[0] < 4:
+                raise Retry
+
+        self.assertEqual(i[0], 4)
+        self.assertEqual(i[1].description, "Nice doc")
+
+    def test_run_pos_tries(self):
+        tm = self._makeOne()
+
+        with self.assertRaises(ValueError):
+            tm.run(0)(lambda : None)
+        with self.assertRaises(ValueError):
+            @tm.run(-1)
+            def _():
+                raise AssertionError("Never called")
+
+    def test_run_stop_on_success(self):
+        import transaction.interfaces
+
+        tm = self._makeOne()
+        i = [0, None]
+
+        @tm.run()
+        def meaning():
+            i[0] += 1
+            i[1] = tm.get()
+            return 43
+
+        self.assertEqual(i[0], 1)
+        self.assertEqual(meaning, 43)
+        self.assertEqual(i[1].description, "meaning")
+
+    def test_run_retries_but_gives_up(self):
+        import transaction.interfaces
+        class Retry(transaction.interfaces.TransientError):
+            pass
+
+        tm = self._makeOne()
+        i = [0]
+
+        with self.assertRaises(Retry):
+            @tm.run()
+            def _():
+                i[0] += 1
+                raise Retry
+
+        self.assertEqual(i[0], 3)
+
+    def test_run_propigates_errors(self):
+        tm = self._makeOne()
+        with self.assertRaises(ValueError):
+            @tm.run
+            def _():
+                raise ValueError
+
+    def test_run_defer_to_dm(self):
+        import transaction.tests.savepointsample
+
+        class DM(transaction.tests.savepointsample.SampleSavepointDataManager):
+            def should_retry(self, e):
+                if 'should retry' in str(e):
+                    return True
+
+        ntry = [0]
+        dm = transaction.tests.savepointsample.SampleSavepointDataManager()
+        dm2 = DM()
+        with transaction.manager:
+            dm2['ntry'] = 0
+
+        @transaction.manager.run
+        def _():
+            ntry[0] += 1
+            dm['ntry'] = ntry[0]
+            dm2['ntry'] = ntry[0]
+            if ntry[0] % 3:
+                raise ValueError('we really should retry this')
+
+        self.assertEqual(ntry[0], 3)
+
+    def test_run_callable_with_bytes_doc(self):
+        import transaction
+        class Callable(object):
+
+            def __init__(self):
+                self.__doc__ = b'some bytes'
+                self.__name__ = b'more bytes'
+
+            def __call__(self):
+                return 42
+
+        result = transaction.manager.run(Callable())
+        self.assertEqual(result, 42)
+
     def test__retryable_w_transient_error(self):
         from transaction.interfaces import TransientError
         tm = self._makeOne()
@@ -255,6 +479,19 @@ class TransactionManagerTests(unittest.TestCase):
         tm.get()._resources.append(_Resource())
         self.assertTrue(tm._retryable(Exception, object()))
 
+    def test__retryable_w_multiple(self):
+        class _Resource(object):
+            _should = True
+            def should_retry(self, err):
+                return self._should
+        tm = self._makeOne()
+        res1 = _Resource()
+        res1._should = False
+        res2 = _Resource()
+        tm.get()._resources.append(res1)
+        tm.get()._resources.append(res2)
+        self.assertTrue(tm._retryable(Exception, object()))
+
     # basic tests with two sub trans jars
     # really we only need one, so tests for
     # sub1 should identical to tests for sub2
@@ -414,6 +651,135 @@ class TransactionManagerTests(unittest.TestCase):
 
         assert nosub1._p_jar.ctpc_abort == 1
 
+    def test_notify_transaction_late_comers(self):
+        # If a datamanager registers for synchonization after a
+        # transaction has started, we should call newTransaction so it
+        # can do necessry setup.
+        import mock
+        from .. import TransactionManager
+        manager = TransactionManager()
+        sync1 = mock.MagicMock()
+        manager.registerSynch(sync1)
+        sync1.newTransaction.assert_not_called()
+        t = manager.begin()
+        sync1.newTransaction.assert_called_with(t)
+        sync2 = mock.MagicMock()
+        manager.registerSynch(sync2)
+        sync2.newTransaction.assert_called_with(t)
+
+        # for, um, completeness
+        t.commit()
+        for s in sync1, sync2:
+            s.beforeCompletion.assert_called_with(t)
+            s.afterCompletion.assert_called_with(t)
+
+    def test_unregisterSynch_on_transaction_manager_from_serparate_thread(self):
+        # We should be able to get the underlying manager of the thread manager
+        # cand call methods from other threads.
+
+        import threading, transaction
+
+        started = threading.Event()
+        stopped = threading.Event()
+
+        synchronizer = self
+
+        class Runner(threading.Thread):
+
+            def __init__(self):
+                threading.Thread.__init__(self)
+                self.manager = transaction.manager.manager
+                self.setDaemon(True)
+                self.start()
+
+            def run(self):
+                self.manager.registerSynch(synchronizer)
+                started.set()
+                stopped.wait()
+
+        runner = Runner()
+        started.wait()
+        runner.manager.unregisterSynch(synchronizer)
+        stopped.set()
+        runner.join(1)
+
+    # The lack of the method below caused a test failure in one run
+    #   -- caused indirectly by the failure of another test (this
+    #   indicates that the tests in this suite are not fully isolated).
+    #   However, defining the methods below reduced the "test coverage"
+    #   once the initial test failure has been fixed.
+    #   We therefore comment them out.
+    ##    # the preceeding test (maybe others as well) registers `self` as
+    ##    #   synchronizer; satisfy the `ISynchronizer` requirements
+    ##    def newTransaction(self, transaction):
+    ##        pass
+    ##
+    ##    beforeCompletion = afterCompletion = newTransaction
+
+
+class TestThreadTransactionManager(unittest.TestCase):
+
+    def test_interface(self):
+        import transaction
+        zope.interface.verify.verifyObject(interfaces.ITransactionManager,
+                                           transaction.manager)
+
+    def test_sync_registration_thread_local_manager(self):
+        import transaction
+
+        sync = mock.MagicMock()
+        sync2 = mock.MagicMock()
+        self.assertFalse(transaction.manager.registeredSynchs())
+        transaction.manager.registerSynch(sync)
+        self.assertTrue(transaction.manager.registeredSynchs())
+        transaction.manager.registerSynch(sync2)
+        self.assertTrue(transaction.manager.registeredSynchs())
+        t = transaction.begin()
+        sync.newTransaction.assert_called_with(t)
+        transaction.abort()
+        sync.beforeCompletion.assert_called_with(t)
+        sync.afterCompletion.assert_called_with(t)
+        transaction.manager.unregisterSynch(sync)
+        self.assertTrue(transaction.manager.registeredSynchs())
+        transaction.manager.unregisterSynch(sync2)
+        self.assertFalse(transaction.manager.registeredSynchs())
+        sync.reset_mock()
+        transaction.begin()
+        transaction.abort()
+        sync.newTransaction.assert_not_called()
+        sync.beforeCompletion.assert_not_called()
+        sync.afterCompletion.assert_not_called()
+
+        self.assertFalse(transaction.manager.registeredSynchs())
+        transaction.manager.registerSynch(sync)
+        transaction.manager.registerSynch(sync2)
+        t = transaction.begin()
+        sync.newTransaction.assert_called_with(t)
+        self.assertTrue(transaction.manager.registeredSynchs())
+        transaction.abort()
+        sync.beforeCompletion.assert_called_with(t)
+        sync.afterCompletion.assert_called_with(t)
+        transaction.manager.clearSynchs()
+        self.assertFalse(transaction.manager.registeredSynchs())
+        sync.reset_mock()
+        transaction.begin()
+        transaction.abort()
+        sync.newTransaction.assert_not_called()
+        sync.beforeCompletion.assert_not_called()
+        sync.afterCompletion.assert_not_called()
+
+    def test_explicit_thread_local_manager(self):
+        import transaction.interfaces
+
+        self.assertFalse(transaction.manager.explicit)
+        transaction.abort()
+        transaction.manager.explicit = True
+        self.assertTrue(transaction.manager.explicit)
+        with self.assertRaises(transaction.interfaces.NoTransaction):
+            transaction.abort()
+        transaction.manager.explicit = False
+        transaction.abort()
+
 
 class AttemptTests(unittest.TestCase):
 
@@ -442,13 +808,13 @@ class AttemptTests(unittest.TestCase):
         self.assertTrue(manager.aborted)
 
     def test___exit__no_exc_abort_exception_after_nonretryable_commit_exc(self):
-        manager = DummyManager(raise_on_abort=ValueError, 
+        manager = DummyManager(raise_on_abort=ValueError,
                                raise_on_commit=KeyError)
         inst = self._makeOne(manager)
         self.assertRaises(ValueError, inst.__exit__, None, None, None)
         self.assertTrue(manager.committed)
         self.assertTrue(manager.aborted)
-        
+
     def test___exit__no_exc_retryable_commit_exception(self):
         from transaction.interfaces import TransientError
         manager = DummyManager(raise_on_commit=TransientError)
@@ -473,13 +839,49 @@ class AttemptTests(unittest.TestCase):
         self.assertRaises(KeyError, inst.__exit__, KeyError, KeyError(), None)
         self.assertFalse(manager.committed)
         self.assertTrue(manager.aborted)
-        
+
+    def test_explicit_mode(self):
+        from .. import TransactionManager
+        from ..interfaces import AlreadyInTransaction, NoTransaction
+
+        tm = TransactionManager()
+        self.assertFalse(tm.explicit)
+
+        tm = TransactionManager(explicit=True)
+        self.assertTrue(tm.explicit)
+        for name in 'get', 'commit', 'abort', 'doom', 'isDoomed', 'savepoint':
+            with self.assertRaises(NoTransaction):
+                getattr(tm, name)()
+
+        t = tm.begin()
+        with self.assertRaises(AlreadyInTransaction):
+            tm.begin()
+
+        self.assertTrue(t is tm.get())
+
+        self.assertFalse(tm.isDoomed())
+        tm.doom()
+        self.assertTrue(tm.isDoomed())
+        tm.abort()
+
+        for name in 'get', 'commit', 'abort', 'doom', 'isDoomed', 'savepoint':
+            with self.assertRaises(NoTransaction):
+                getattr(tm, name)()
+
+        t = tm.begin()
+        self.assertFalse(tm.isDoomed())
+        with self.assertRaises(AlreadyInTransaction):
+            tm.begin()
+        tm.savepoint()
+        tm.commit()
+
+
 
 class DummyManager(object):
     entered = False
     committed = False
     aborted = False
-    
+
     def __init__(self, raise_on_commit=None, raise_on_abort=None):
         self.raise_on_commit = raise_on_commit
         self.raise_on_abort = raise_on_abort
@@ -487,7 +889,7 @@ class DummyManager(object):
     def _retryable(self, t, v):
         from transaction._manager import TransientError
         return issubclass(t, TransientError)
-        
+
     def __enter__(self):
         self.entered = True
 
@@ -495,7 +897,7 @@ class DummyManager(object):
         self.aborted = True
         if self.raise_on_abort:
             raise self.raise_on_abort
-        
+
     def commit(self):
         self.committed = True
         if self.raise_on_commit:
@@ -522,7 +924,7 @@ class TestTxnException(Exception):
     pass
 
 
-class BasicJar:
+class BasicJar(object):
 
     def __init__(self, errors=(), tracing=0):
         if not isinstance(errors, tuple):
@@ -549,7 +951,7 @@ class BasicJar:
         return self.__class__.__name__
 
     def check(self, method):
-        if self.tracing:
+        if self.tracing: # pragma: no cover
             print('%s calling method %s'%(str(self.tracing),method))
 
         if method in self.errors:
@@ -595,14 +997,8 @@ def positive_id(obj):
     _ADDRESS_MASK = 256 ** struct.calcsize('P')
 
     result = id(obj)
-    if result < 0:
+    if result < 0: # pragma: no cover
+        # Happens...on old 32-bit systems?
         result += _ADDRESS_MASK
         assert result > 0
     return result
-
-
-def test_suite():
-    return unittest.TestSuite((
-        unittest.makeSuite(TransactionManagerTests),
-        unittest.makeSuite(AttemptTests),
-    ))
